//==========================================================================
// File: traversal.c
// Description: Functions for enumerating vertices, edges, and faces
//   of subdivisions
// Created: Wed Dec  1 18:58:35 EST 1999 by Pat Morin
//==========================================================================
#include <LEDA/segment.h>
#include <LEDA/ray.h>
#include <LEDA/stack.h>
#include <LEDA/face_array.h>

#include <LEDA/window.h>
#include <LEDA/color.h>

#include "traversal.h"
#include "report.h"

// return the minimum of a and b
#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

// return the smaller angle between two directions
#define DIR_CMP(d1, d2) min(fabs((d1)-(d2)), fabs(fabs((d1)-(d2))-PI_TIMES_2))

// see comment in header
#ifdef MEASURING
int bm_cmp_count;
int bkoo_cmp_count;
#endif //MEASURING


//=======================================================================
// Compute a point p very close to the midpoint of s and t
//=======================================================================
void get_reference_point(const POINT &s, const POINT &t, POINT &p)
{
  POINT c((s.to_vector() + t.to_vector())/2);
  vector v = (s.to_vector()-t.to_vector()).rotate90();
  v = v.norm();
  v = v * 0.0001;
  p = POINT(c.to_vector()+v);
}

//=======================================================================
// Measure the angle p1,p2,p3 as the smaller of the clockwise and 
// counterclockwise angles.
//=======================================================================
inline 
double angle_measure(const point &p1, const point &p2, const point &p3)
{
  double dir1 = ray(p2, p1).direction();
  double dir2 = ray(p2, p3).direction();
  return DIR_CMP(dir1, dir2);
}

//=======================================================================
// Compare two edges using the \prec_p relation
//=======================================================================
inline 
int edge_compare(const POINT &s1, const POINT &t1, 
		 const POINT &s2, const POINT &t2, const POINT &p) 
{
  point px1, px2;
  int val1 = robust_distance_test(SEGMENT(s1, t1), p, px1);
  int val2 = robust_distance_test(SEGMENT(s2, t2), p, px2);

  double dist1 = px1.sqr_dist(p);
  double dist2 = px2.sqr_dist(p);

  // discrimate based on distance to p
  if (dist1 < dist2)
    return -1;
  if (dist1 == dist2) {
    // discriminate based on contact point with circle
    double dir1 = ray(p, px1).direction();
    double dir2 = ray(p, px2).direction();
    if (dir1 < dir2)
      return -1;
    if (dir1 == dir2) {
      // discriminate based on angles
      point pt1 = (val1 == D_SOURCE) ? t1 : s1;
      point pt2 = (val2 == D_SOURCE) ? t2 : s2;
      double ang1 = angle_measure(p, px1, pt1);
      double ang2 = angle_measure(p, px2, pt2);
      if (ang1 < ang2)
	return -1;
      if (ang1 == ang2) {
	double diro = ray(px1, p).direction();
	double a1 = ray(px1, pt1).direction();
	a1 = (a1 > diro) ? a1 - diro : a1 + PI_TIMES_2 - diro;
	double a2 = ray(px2, pt2).direction();
	a2 = (a2 > diro) ? a2 - diro : a2 + PI_TIMES_2 - diro;
	if (a1 < a2)
	  return -1;
	if (a1 == a2)
	  return 0;
      }
    }
  }
  return 1;
}

//=======================================================================
// Test if the edge e is the entry edge of G.face_of(e) using the 
// de Berg et al. test
//=======================================================================
bool test_entry_bkoo(const GRAPH<point, int> &G, edge e, const point &p)
{
  edge estart = e;
  e = G.face_cycle_succ(e);
  while (e != estart) {
#ifdef MEASURING
    bkoo_cmp_count++;
#endif // MEASURING
    if (edge_compare(G[G.source(estart)], G[G.target(estart)],
		     G[G.source(e)], G[G.target(e)], p) >= 0) {
      return false;
    }
    e = G.face_cycle_succ(e);
  }
  return true;
}

//=======================================================================
// Test if the edge e is the entry edge of G.face_of(e) using the 
// Bose and Morin test
//=======================================================================
bool test_entry_bm(const GRAPH<point, int> &G, edge e, const point &p)
{
  edge estart = e;
  edge epred = G.face_cycle_pred(e);
  edge esucc = G.face_cycle_succ(e);
  while (1 < 2) {
#ifdef MEASURING
    bm_cmp_count++;
#endif // MEASURING
    if (edge_compare(G[G.source(estart)], G[G.target(estart)],
		     G[G.source(esucc)], G[G.target(esucc)], p) >= 0)
      return false;
    if (esucc == epred)
      break;
#ifdef MEASURING
    bm_cmp_count++;
#endif // MEASURING
    if (edge_compare(G[G.source(estart)], G[G.target(estart)],
		     G[G.source(epred)], G[G.target(epred)], p) >= 0)
      return false;
    esucc = G.face_cycle_succ(esucc);
    if (esucc == epred)
      break;
    epred = G.face_cycle_pred(epred);
  }
  return true;
}

//=======================================================================
// Traverse the subdivision G beginning at the edge estart using de Berg
// et al's constant-additional-memory subdivision traversal algorithm.
//=======================================================================
void bkoo_traversal(const GRAPH<POINT, int> &G, edge estart) 
{
  node v;
  edge e;
  window W(600, 600, "unit graph debug window");  
  W.set_bg_color(ivory);
  W.open(window::center,window::center);
  double xmin = HUGE_VAL, xmax = -HUGE_VAL, 
    ymin = HUGE_VAL, ymax = -HUGE_VAL;
  forall_nodes (v, G) {
    if (G[v].xcoord() < xmin) xmin = G[v].xcoord();
    if (G[v].xcoord() > xmax) xmax = G[v].xcoord();
    if (G[v].ycoord() < ymin) ymin = G[v].ycoord();
    if (G[v].ycoord() > ymax) ymax = G[v].ycoord();
  }
  W.init(xmin, xmax, ymin);
  W.clear();
  forall_edges (e, G) {
    W.draw_segment(G[G.source(e)], G[G.target(e)], 
		   black);
  }

  POINT p;
  get_reference_point(G[G.source(estart)], G[G.target(estart)], p);

  e = estart;
  do {
    W.draw_segment(G[G.source(e)], G[G.target(e)], red);
    W.draw_segment(G[G.source(estart)], G[G.target(estart)], green);
    // report e if necessary
    point s(G[G.source(e)]);
    point t(G[G.target(e)]);
    double ds = s.sqr_dist(p);
    double dt = t.sqr_dist(p);
    if (ds < dt || (ds == dt && ray(p,s).direction() < ray(p,t).direction())) {
      report(e);
    }
    // report G.target(e) if necessary
    point u(G[G.target(G.face_cycle_succ(e))]);
    if ((right_turn(p, t, s) && !right_turn(p, t, u)) ||
	(right_turn(p, t, s) && right_turn(p, t, u) && right_turn(s, t, u)) ||
	(!right_turn(p, t, s) && !right_turn(p, t, u) && right_turn(s, t, u))){
      report(G.target(e));
    }

    if (test_entry_bkoo(G, e, p)) {
      // return to parent
      W.draw_segment(G[G.source(e)], G[G.target(e)], blue);
      report(G.face_of(e));
      e = G.reversal(e);
    } else if (test_entry_bkoo(G, G.reversal(e), p)) {
      // visit child
      W.draw_segment(G[G.source(e)], G[G.target(e)], green);
      e = G.reversal(e);
    }
    e = G.face_cycle_succ(e);
  } while (e != estart);
  return;
}

//=======================================================================
// Traverse the subdivision G beginning at the edge estart using 
// Bose and Morin's constant-additional-memory subdivision traversal 
// algorithm.
//=======================================================================
void bm_traversal(const GRAPH<POINT, int> &G, edge estart) 
{
  POINT p;
  get_reference_point(G[G.source(estart)], G[G.target(estart)], p);

  edge e;

  e = estart;
  do {
    // report e if necessary
    point s(G[G.source(e)]);
    point t(G[G.target(e)]);
    double ds = s.sqr_dist(p);
    double dt = t.sqr_dist(p);
    if (ds < dt || (ds == dt && ray(p,s).direction() < ray(p,t).direction())) {
      report(e);
    }

    // report G.target(e) if necessary
    point u(G[G.target(G.face_cycle_succ(e))]);
    if ((right_turn(p, t, s) && !right_turn(p, t, u)) ||
	(right_turn(p, t, s) && right_turn(p, t, u) && right_turn(s, t, u)) ||
	(!right_turn(p, t, s) && !right_turn(p, t, u) && right_turn(s, t, u))){
      report(G.target(e));
    }
    if (test_entry_bm(G, e, p)) {
      // return to parent
      report(G.face_of(e));
      e = G.reversal(e);
    } else if (test_entry_bm(G, G.reversal(e), p)) {
      // visit child
      e = G.reversal(e);
    }
    e = G.face_cycle_succ(e);
  } while (e != estart);
  return;
}


//=======================================================================
// Traverse the subdivision G beginning at the edge estart using depth
// first search on the faces of G.
//=======================================================================
void dfs_traversal(const GRAPH<POINT, int> &G, edge estart)
{
  face_array<bool> visited(G);
  stack<face> S;

  // first report the faces
  face f = G.face_of(estart);
  S.push(f);

  while (!S.empty()) {
    f = S.pop();
    if (!visited[f]) {
      visited[f] = true;
      report(f);
      edge e;
      forall_face_edges (e, f) {
	face f2 = G.face_of(G.reversal(e));
	if (!visited[f2]) {
	  S.push(f2);
	}
      }
    }
  }

  // next the vertices and edges
  // do a DFS and report an edge when it is a back edge
  node_array<bool> visited2(G);
  stack<node> S2;
  node v = G.source(estart);
  S2.push(v);

  while (!S2.empty()) {
    v = S2.pop();
    if (!visited2[v]) {
      visited2[v] = true;
      report(v);
      edge e;
      forall_adj_edges (e, v) {
	if (visited2[G.target(e)]) {
	  report(e);
	} else {
	  S2.push(G.target(e));
	} 
      }
    }
  }
}


